# -*- coding: utf-8 -*-
'''
National Instruments SystemLink Assets' Health Monitoring Beacon
'''
# Import Python libs
import logging
import os
import sys
from typing import Any, Dict, List, Tuple, Union

# Import salt libs
# pylint: disable=import-error,3rd-party-module-not-gated
import salt.utils

__virtualname__ = 'ni_asset_health_monitoring'

# Set up logging
log = logging.getLogger(__name__)

# For firing local events
EVENT = None

BEACON_INITIALIZED = False
BEACON_INTERVAL_DISCONNECTED = 15
BEACON_INTERVAL_DEFAULT = 300
BEACON_DEFAULT_ENABLED = True

MONITORING_ENABLED_KEY = 'asset_health_monitoring_enabled'
MONITORING_INTERVAL_KEY = 'asset_health_monitoring_interval'

# Placeholder for the _ni_asset_util_module
ni_asset_health_util = None  # pylint: disable=invalid-name


def _import_modules():
    # Add extra search paths to load 'ni_asset_util'
    paths_to_add = __opts__['module_dirs']
    extmods_module_dir = os.path.join(__opts__['extension_modules'], 'modules')
    if extmods_module_dir not in paths_to_add:
        paths_to_add.append(extmods_module_dir)
    sys.path.extend(paths_to_add)
    global ni_asset_health_util  # pylint: disable=global-statement
    try:
        import _ni_asset_health_util as ni_asset_health_util
    finally:
        # Remove the extra search paths that we added to sys.path
        for path in paths_to_add:
            sys.path.remove(path)


def __virtual__() -> Union[str, Tuple[bool, str]]:
    '''
    During module lazy loading, will return the virtual name of this module.

    :return: The virtual name of this module. On error, return a 2-tuple with
        ``False`` for the first item and the error message for the second.
    '''
    _import_modules()
    return __virtualname__


def validate(config: List[Dict[str, Any]]) -> Tuple[bool, str]:
    '''
    Validate the beacon configuration

    :param config: The beacon configuration.
    :return: A 2-tuple. The first item is ``True`` if the
        configuration is valid and ``False`` otherwise. The second item
        is a message.
    '''
    if not isinstance(config, list):
        return False, ('Configuration for ni_asset_health_monitoring '
                       'beacon must be a list.')
    return True, 'Valid beacon configuration'


def beacon(config: List[Dict[str, Any]]) -> List[Dict[str, Any]]:  # pylint: disable=unused-argument
    '''
    National Instruments SystemLink Assets' Health Monitoring Beacon

    :param config: The beacon configuration.
    :return: A list of events to send to the master.
    '''
    global BEACON_INITIALIZED  # pylint: disable=global-statement

    try:
        if not BEACON_INITIALIZED:
            success = _init_beacon()
            if not success:
                return []
        ni_asset_health_util.refresh_pxi_chassis()
        _connected_beacon_interval()
    except Exception as exc:  # pylint: disable=broad-except
        log.error(
            'Unexpected exception in "ni_asset_health_monitoring beacon": %s',
            exc, exc_info=True
        )
        _disconnected_beacon_interval()

    return []


def _init_beacon() -> bool:
    '''
    Initialize the beacon

    :return: ``True`` if initialized successfully, ``False`` otherwise.
    '''
    global BEACON_INITIALIZED  # pylint: disable=global-statement

    enabled = __grains__.get(MONITORING_ENABLED_KEY, BEACON_DEFAULT_ENABLED)
    if __opts__.get('master_type') == 'disable' or not enabled:
        _fire_local_event(
            {
                'func': 'disable_beacon',
                'name': __virtualname__
            },
            'manage_beacons'
        )
        return False

    BEACON_INITIALIZED = True
    return True


def _fire_local_event(event_data: Dict[str, Any], event_tag: str) -> None:
    '''
    Fire an event to the local event bus.

    :param event_data: The data associated with the event.
    :param event_tag: The identifier of the event.
    '''
    global EVENT  # pylint: disable=global-statement

    if not EVENT:
        EVENT = salt.utils.event.get_event(
            'minion', opts=__opts__, listen=False
        )
    EVENT.fire_event(event_data, event_tag)


def _connected_beacon_interval() -> None:
    '''
    This is invoked once we successfully connected
    and want to ensure that the long term beacon
    interval is in effect.
    '''
    global BEACON_INTERVAL_DEFAULT

    interval = __grains__.get(MONITORING_INTERVAL_KEY, BEACON_INTERVAL_DEFAULT)
    current_interval = __opts__['beacons'][__virtualname__][0].get('interval', BEACON_INTERVAL_DEFAULT)
    if current_interval != interval:
        _set_beacon_interval(interval)


def _disconnected_beacon_interval() -> None:
    '''
    This is invoked when we have not successfully connected
    or have been disconnected due to error and want to ensure
    that the short term beacon interval is in effect.
    '''
    global BEACON_INITIALIZED
    global BEACON_INTERVAL_DEFAULT

    current_interval = __opts__['beacons'][__virtualname__][0].get('interval', BEACON_INTERVAL_DEFAULT)
    if current_interval != BEACON_INTERVAL_DISCONNECTED:
        _set_beacon_interval(BEACON_INTERVAL_DISCONNECTED)


def _set_beacon_interval(interval: int) -> None:
    '''
    This is invoked once we successfully connected
    and want to ensure that the long term beacon
    interval is in effect.

    :param interval: The interval on which to run the beacon.
    '''
    _fire_local_event(
        {
            'func': 'modify',
            'name': __virtualname__,
            'beacon_data': [{
                'interval': interval
            }]
        },
        'manage_beacons'
    )
