# -*- coding: utf-8 -*-
"""
National Instruments SystemLink Assets' Health Utils Module
"""
# Import Python libs
import asyncio
import json
import logging
import os
import sys
from pathlib import Path
from typing import Any, Dict, List, Optional
from urllib.parse import quote
from datetime import datetime

from asset_discovery.models import (
    Asset,
    AssetType,
    BusType,
    CoolingHealth,
    FanMode,
    CoolingProfile,
    SensorHealthState,
    LifetimePowerOn,
    OperationalState,
    SensorModel,
    SensorCollection,
    FanSensorModel,
    SensorProperties,
    AssetHealthData,
    OperationalStateValue)
from asset_discovery.repository.path_resolver import PathResolver

# pylint: enable=import-error,3rd-party-module-not-gated

# Import National Instruments libs
# pylint: disable=import-error,3rd-party-module-not-gated
try:
    # noinspection PyUnresolvedReferences
    from systemlink.messagebus.paths import get_http_master_file
    # noinspection PyUnresolvedReferences
    from systemlink.clientconfig import get_configuration_by_id
    # noinspection PyUnresolvedReferences
    from systemlink.clientconfig.configuration import Configuration
    # noinspection PyUnresolvedReferences
    from systemlink.clients.nitag import (
        ApiClient, ApiException, Tag, TagListAndMergeFlag,
        TagsApi, TagUpdate, TagValue, TimestampedTagValue
    )
    # noinspection PyUnresolvedReferences
    from systemlink.clients.nitag.rest import RESTResponse

    HAS_SYSTEMLINK = True
except ImportError:
    HAS_SYSTEMLINK = False

# Import local libs
# This file may be loaded out of __pycache__, so the
# directory of its .py may not be in the search path.
IMPORT_PATH = os.path.dirname(__file__)
if IMPORT_PATH.endswith('__pycache__'):
    IMPORT_PATH = os.path.dirname(IMPORT_PATH)
sys.path.append(IMPORT_PATH)
try:
    import _ni_asset_util
finally:
    # Remove the extra search path that we added to sys.path
    sys.path.remove(IMPORT_PATH)
# pylint: disable=import-error,3rd-party-module-not-gated

# Set up logging
log = logging.getLogger(__name__)

HTTP_MASTER_CONFIGURATION_ID = 'SYSTEMLINK_MASTER'

DEFAULT_ASSET_HEALTH_MONITORING_ENABLED = True
DEFAULT_ASSET_HEALTH_MONITORING_INTERVAL = 300

HEALTH = 'Health'
TEMPERATURE = 'Temperature'
VOLTAGE = 'Voltage'
POWER = 'Power'
COOLING = 'Cooling'
LIFETIME_POWER_ON = 'TotalPowerOnHours'
OPERATIONAL_STATE = 'OperationalState'
FAN_MODE = 'Mode'
FAN_USER_RPM = 'User'
COOLING_PROFILE = 'Profile'
COOLING_OPERATIONAL_STATE = 'OperationalState'
READING = 'Reading'
ALARM = 'Alarm'
STATE = 'State'
FAN = 'Fan'

TYPE_DOUBLE = 'DOUBLE'
TYPE_INT = 'INT'
TYPE_STRING = 'STRING'

UNITS_CELSIUS = '°C'
UNITS_VOLTS = 'V'
UNITS_WATTS = 'W'
UNITS_RPM = 'RPM'
UNITS_HOURS = 'hours'
UNITS_CUSTOM = 'Custom'

PROPS_SENSOR_NAME = 'Sensor name'
PROPS_SYSTEM_NAME = 'System name'
PROPS_VENDOR_NAME = 'Vendor name'
PROPS_MODEL_NAME = 'Model name'
PROPS_SERIAL_NUMBER = 'Serial number'
PROPS_ASSET_NAME = 'Asset name'
PROPS_OPERATIONAL_STATE_STATUS = 'Status'
PROPS_SUPPORTED_MODES = 'Supported modes'
PROPS_MANUAL_RPM_UPPER_BOUND = 'Upper RPM bound'
PROPS_MANUAL_RPM_LOWER_BOUND = 'Lower RPM bound'
PROPS_STATE_NAME = 'State name'
PROPS_ALARM_LEVEL = 'Alarm level'

SENSOR_PROPS_SERIALIZATION_KEYS = {
    SensorProperties.Interval: 'Polling interval',
    SensorProperties.LowerUser: 'Lower user threshold',
    SensorProperties.UpperUser: 'Upper user threshold',
    SensorProperties.LowerRisk: 'Lower damage risk threshold',
    SensorProperties.UpperRisk: 'Upper damage risk threshold',
    SensorProperties.LowerCritical: 'Lower critical threshold',
    SensorProperties.UpperCritical: 'Upper critical threshold',
    SensorProperties.Accuracy: 'Accuracy',
    SensorProperties.Nominal: 'Nominal'
}

COOLING_HEALTH_COOLING_PROFILE_VALUES = {
    CoolingProfile.PROFILE_38_W: 38,
    CoolingProfile.PROFILE_58_W: 58,
    CoolingProfile.PROFILE_82_W: 82,
    CoolingProfile.PROFILE_CUSTOM: -1
}

SENSOR_HEALTH_STATE_VALUES = {
    SensorHealthState.UNKNOWN: -1,
    SensorHealthState.NORMAL: 0,
    SensorHealthState.EXCEEDS_USER_DEFINED_THRESHOLD: 1,
    SensorHealthState.OUT_OF_SPEC: 2,
    SensorHealthState.RISK_OF_DAMAGE: 3
}

RESERVED_RETENTION_TYPE_KEY = 'nitagRetention'
RESERVED_RETENTION_DAYS_KEY = 'nitagHistoryTTLDays'
RESERVED_UNITS_KEY = 'units'
RESERVED_HYPERLINK_KEY = 'hyperlink'
RESERVED_DISPLAY_NAME_KEY = 'displayName'
RESERVED_MINION_ID = 'minionId'

RETENTION_DURATION_DAYS = '365'
RETENTION_TYPE = 'DURATION'

DATA_DIRECTORY = 'Data'
TAGS_DIRECTORY = 'Tags'
LAST_UPDATED_TAGS_TIMESTAMPS_FILE = 'LastUpdatedTimestamps.json'


class TagWrapper:
    def __init__(self, create_or_update_model: Tag, update_value_model: TagUpdate):
        self._create_or_update_model = create_or_update_model
        self._update_value_model = update_value_model

    @property
    def create_or_update_model(self):
        return self._create_or_update_model

    @property
    def update_value_model(self):
        return self._update_value_model


def refresh_pxi_chassis() -> None:
    chassis = _get_pxi_chassis()
    chassis_identifications = [asset.identification.to_dict() for asset in chassis]
    _ni_asset_util.get_asset_discovery_engine().refresh(assets=chassis_identifications)


def _get_pxi_chassis() -> List[Asset]:
    all_assets = _ni_asset_util.get_asset_discovery_engine().list_assets()
    return _filter_pxi_chassis(all_assets)


def _filter_pxi_chassis(assets: List[Asset]) -> List[Asset]:
    return [asset for asset in assets if asset.asset_type == AssetType.CHASSIS and asset.provided_bus_type == BusType.PCI_PXI]


def create_and_publish_tags(
        assets: List[Asset],
        workspace: str,
        hostname: str,
        minion_id: str) -> None:
    global HAS_SYSTEMLINK  # pylint: disable=global-statement

    if not HAS_SYSTEMLINK:
        raise Exception("Failed to import SystemLink dependencies.")
    assets_with_health_data = _get_assets_with_health_data(assets)
    tags = _create_tags(workspace=workspace, assets_with_health_data=assets_with_health_data, hostname=hostname, minion_id=minion_id)
    if tags is None or len(tags) == 0:
        return
    create_or_update_models = [tag.create_or_update_model for tag in tags]
    tags_and_merge_flag = TagListAndMergeFlag(create_or_update_models, False)
    tag_value_update_models = [tag.update_value_model for tag in tags]
    _publish_tags_with_loop(
        tags_and_merge_flag=tags_and_merge_flag,
        tag_updates=tag_value_update_models)


def _get_assets_with_health_data(assets: List[Asset]) -> List[Asset]:
    """
    Gets assets with health data.
    """
    assets_with_health = []
    for asset in assets:
        if asset.health is not None:
            assets_with_health.append(asset)
    return assets_with_health


def _publish_tags_with_loop(tags_and_merge_flag: TagListAndMergeFlag, tag_updates: List[TagUpdate]):
    loop = None
    new_event_loop_created = None
    try:
        new_event_loop_created = _ensure_event_loop()
        loop = asyncio.get_event_loop()
        if not loop.is_running():
            loop.run_until_complete(_publish_tags_over_http(tags_and_merge_flag, tag_updates))
        else:
            asyncio.ensure_future(_publish_tags_over_http(tags_and_merge_flag, tag_updates))
    finally:
        if new_event_loop_created is True and loop is not None:
            loop.stop()
            loop.close()
            asyncio.set_event_loop(None)


def _ensure_event_loop() -> bool:
    try:
        asyncio.get_event_loop()
        return False
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        return True


async def _publish_tags_over_http(tags_and_merge_flag: TagListAndMergeFlag, tag_updates: List[TagUpdate]):
    writer = TagHttpWriter()
    if writer.get_configuration() is None:
        log.warning("Failed to create the TagHttpWriter with a valid config.")
        return
    await writer.create_or_update_tags(tags_and_merge_flag)
    await writer.update_tag_values(tag_updates)


def _create_tags(assets_with_health_data: List[Asset], workspace: str, hostname: str, minion_id: str) -> List[TagWrapper]:
    tags = []
    common_properties = _get_common_properties(hostname=hostname, minion_id=minion_id)
    tags_last_updated_timestamps = _read_tag_last_updated_timestamp_file()
    for asset in assets_with_health_data:
        tags_for_asset = _create_asset_health_tags(
            asset=asset,
            common_properties=common_properties,
            workspace=workspace,
            tags_last_updated_timestamps=tags_last_updated_timestamps)
        tags.extend(tags_for_asset)
    return tags


def _get_common_properties(hostname: str, minion_id: str) -> Dict[str, str]:
    return {
        PROPS_SYSTEM_NAME: hostname,
        RESERVED_MINION_ID: minion_id,
        RESERVED_RETENTION_TYPE_KEY: RETENTION_TYPE,
        RESERVED_RETENTION_DAYS_KEY: RETENTION_DURATION_DAYS
    }


def _create_asset_health_tags(
        asset: Asset,
        common_properties: Dict[str, str],
        workspace: str,
        tags_last_updated_timestamps: Dict[str, str]) -> List[TagWrapper]:
    asset_properties = _get_asset_properties(asset)
    base_properties = {**common_properties, **asset_properties}
    path_prefix = _get_asset_health_path_prefix(asset)
    return _create_asset_sensor_tags(
        asset_health_data=asset.health,
        base_properties=base_properties,
        path_prefix=path_prefix,
        workspace=workspace,
        tags_last_updated_timestamps=tags_last_updated_timestamps)


def _get_asset_properties(asset: Asset) -> Dict[str, str]:
    props = {}

    vendor = asset.vendor_name
    if vendor is not None:
        props[PROPS_VENDOR_NAME] = vendor

    model = asset.model_name
    if model is not None:
        props[PROPS_MODEL_NAME] = model

    serial_number = asset.serial_number
    props[PROPS_SERIAL_NUMBER] = serial_number

    props[PROPS_ASSET_NAME] = asset.name
    return props


def _get_asset_health_path_prefix(asset: Asset) -> str:
    vendor = asset.vendor_name
    if vendor is None:
        vendor = str(asset.vendor_number)

    model = asset.model_name
    if model is None:
        model = str(asset.model_number)

    serial_number = str(asset.serial_number)
    vendor = _sanitize_path_component(vendor)
    model = _sanitize_path_component(model)
    serial_number = _sanitize_path_component(serial_number)
    return "Assets.{0}.{1}.{2}.{3}".format(vendor, model, serial_number, HEALTH)


def _create_asset_sensor_tags(
        asset_health_data: AssetHealthData,
        base_properties: Dict[str, str],
        path_prefix: str,
        workspace: str,
        tags_last_updated_timestamps: Dict[str, str]) -> List[TagWrapper]:
    """
    Creates tags that have a newer timestamp than the last recorded one.
    """
    tags = []
    new_temperature_tags = []
    if asset_health_data.temperature_sensors is not None:
        temperature_tags = _create_temp_tags(
            temp_sensors_collection=asset_health_data.temperature_sensors,
            base_properties=base_properties,
            path_prefix=path_prefix,
            workspace=workspace)
        new_temperature_tags = _get_newer_tag(tags_last_updated_timestamps, temperature_tags)

    new_voltage_tags = []
    if asset_health_data.voltage_sensors is not None:
        voltage_tags = _create_volt_tags(
            volt_sensors_collection=asset_health_data.voltage_sensors,
            base_properties=base_properties,
            path_prefix=path_prefix,
            workspace=workspace)
        new_voltage_tags = _get_newer_tag(tags_last_updated_timestamps, voltage_tags)

    new_power_tags = []
    if asset_health_data.power_sensors is not None:
        power_tags = _create_power_tags(
            power_sensors_collection=asset_health_data.power_sensors,
            base_properties=base_properties,
            path_prefix=path_prefix,
            workspace=workspace)
        new_power_tags = _get_newer_tag(tags_last_updated_timestamps, power_tags)

    new_cooling_tags = []
    if asset_health_data.cooling_health is not None:
        cooling_tags = _create_cooling_tags(
            cooling_health=asset_health_data.cooling_health,
            base_properties=base_properties,
            path_prefix=path_prefix,
            workspace=workspace)
        new_cooling_tags = _get_newer_tag(tags_last_updated_timestamps, cooling_tags)

    new_lifetime_power_on_tags = []
    if asset_health_data.lifetime_power_on is not None:
        lifetime_power_on_tags = _create_lifetime_power_on_tags(
            lifetime_power_on=asset_health_data.lifetime_power_on,
            base_properties=base_properties,
            path_prefix=path_prefix,
            workspace=workspace)
        new_lifetime_power_on_tags = _get_newer_tag(tags_last_updated_timestamps, lifetime_power_on_tags)

    new_operational_state_tags = []
    if asset_health_data.operational_state is not None:
        operational_state_tags = _create_operational_state_tags(
            operational_state=asset_health_data.operational_state,
            base_properties=base_properties,
            path_prefix=path_prefix,
            workspace=workspace)
        new_lifetime_power_on_tags = _get_newer_tag(tags_last_updated_timestamps, operational_state_tags)

    tags.extend(new_temperature_tags)
    tags.extend(new_voltage_tags)
    tags.extend(new_power_tags)
    tags.extend(new_cooling_tags)
    tags.extend(new_lifetime_power_on_tags)
    tags.extend(new_operational_state_tags)

    return tags


def _get_newer_tag(tags_last_updated_timestamps: Dict[str, str], tags: List[TagWrapper]):
    return [tag for tag in tags if _is_tag_newer(tags_last_updated_timestamps, tag)]


def _is_tag_newer(tags_last_updated_timestamps: Dict[str, str], tag: TagWrapper) -> bool:
    """
    This method takes in a dictionary with tag paths as keys and last updated timestamps as values and
    compares it with a given tag wrapper.
    If the tag wrapper contains a tag update that has a timestamp newer than the one from the dict, this method
    will return true.
    The timestamps are compared in string iso formats.
    """
    return (_get_tag_timestamp_from_tag_update(tag.update_value_model)
            > (tags_last_updated_timestamps.get(tag.update_value_model.path) or ''))


def _create_temp_tags(
        temp_sensors_collection: SensorCollection,
        base_properties: Dict[str, str],
        path_prefix: str,
        workspace: str) -> List[TagWrapper]:
    path_prefix = "{0}.{1}".format(path_prefix, TEMPERATURE)
    if not SensorCollection.has_any_readings(temp_sensors_collection):
        return []
    tags = []
    for sensor in temp_sensors_collection.sensors:
        tags.extend(_create_temp_sensor_tags(
            sensor=sensor,
            base_properties=base_properties,
            path_prefix=path_prefix,
            workspace=workspace,
            timestamp=temp_sensors_collection.timestamp))
    return tags


def _create_temp_sensor_tags(
        sensor: SensorModel,
        base_properties: Dict[str, str],
        path_prefix: str,
        workspace: str,
        timestamp: datetime) -> List[TagWrapper]:
    reading_tag = _create_temp_reading_tag(
        sensor=sensor,
        base_properties=base_properties,
        path_prefix=path_prefix,
        workspace=workspace,
        timestamp=timestamp)
    alarm_tag = _create_temp_alarm_tag(
        sensor=sensor,
        base_properties=base_properties,
        path_prefix=path_prefix,
        workspace=workspace,
        timestamp=timestamp)
    state_tag = _create_temp_state_tag(
        sensor=sensor,
        base_properties=base_properties,
        path_prefix=path_prefix,
        workspace=workspace,
        timestamp=timestamp)
    sensor_tags = [reading_tag, alarm_tag, state_tag]
    return [t for t in sensor_tags if t is not None]


def _create_temp_reading_tag(
        sensor: SensorModel,
        base_properties: Dict[str, str],
        path_prefix: str,
        workspace: str,
        timestamp: datetime) -> Optional[TagWrapper]:
    value = sensor.reading
    if value == -1:
        return None
    optional_sensor_props_keys = [
        SensorProperties.UpperRisk,
        SensorProperties.LowerRisk,
        SensorProperties.UpperCritical,
        SensorProperties.LowerCritical,
        SensorProperties.UpperUser,
        SensorProperties.LowerUser,
        SensorProperties.Accuracy
    ]
    path_prefix = "{0}.{1}".format(path_prefix, READING)
    units = UNITS_CELSIUS
    tag_type = TYPE_DOUBLE
    display_name = _get_sensor_reading_display_name(sensor.name, base_properties)
    return _create_sensor_tag(
        sensor=sensor,
        value=value,
        tag_type=tag_type,
        display_name=display_name,
        base_properties=base_properties,
        path_prefix=path_prefix,
        optional_props_keys=optional_sensor_props_keys,
        units=units,
        workspace=workspace,
        timestamp=timestamp)


def _get_sensor_reading_display_name(sensor_name: str, base_properties: Dict[str, str]) -> str:
    asset_name = base_properties[PROPS_ASSET_NAME]
    serial_number = base_properties[PROPS_SERIAL_NUMBER]
    return "{0} ({1}) - {2}".format(asset_name, serial_number, sensor_name)


def _create_sensor_tag(
        sensor: Any,
        value: Optional[Any],
        tag_type: str,
        display_name: str,
        base_properties: Dict[str, Any],
        path_prefix: str,
        optional_props_keys: List[str],
        units: Optional[str],
        workspace: str,
        timestamp: datetime) -> Optional[TagWrapper]:
    if value is None or timestamp is None:
        return None
    value = str(value)
    sensor_name = _sanitize_path_component(sensor.name)
    path = "{0}.{1}".format(path_prefix, sensor_name)
    hyperlink = _get_hyperlink(path=path, workspace=workspace)
    sensor_properties = _get_sensor_properties(sensor, optional_props_keys)
    props = {**base_properties, **sensor_properties,
             RESERVED_DISPLAY_NAME_KEY: display_name, RESERVED_HYPERLINK_KEY: hyperlink}
    if units is not None:
        props[RESERVED_UNITS_KEY] = units
    return _create_tag_wrapper(path, value, props, tag_type, timestamp, True)


def _get_hyperlink(path: str, workspace: str) -> str:
    """
    Return a relative URL to the tag details page for the given tag.

    :param path: The tag path to generate a relative URL for.
    :param workspace: The workspace of the system.
    :return: A relative URL to the tag details page for the given tag.
    """
    return '#tagviewer/tag/' + quote(workspace, safe='') + '/' + quote(path, safe='')


def _get_sensor_properties(sensor: SensorModel, optional_props_keys: List[str]) -> Dict[str, Any]:
    props = {
        PROPS_SENSOR_NAME: sensor.name
    }
    for key in optional_props_keys:
        if key in sensor.properties:
            serialization_key = SENSOR_PROPS_SERIALIZATION_KEYS[key]
            value = sensor.properties[key]
            props[serialization_key] = str(value)
    return props


def _create_tag_wrapper(path: str, value: str, properties: Dict[str, Any], tag_type: str, timestamp: datetime, collect_aggregates: bool) -> TagWrapper:
    tag_create_or_update_model = Tag(type=tag_type, properties=properties, path=path, collect_aggregates=collect_aggregates)
    # A timestamp of ``None`` means use the server time.
    tag_updated_value = TimestampedTagValue(value=TagValue(value=value, type=tag_type), timestamp=timestamp)
    tag_update_model = TagUpdate(path=path, updates=[tag_updated_value])
    return TagWrapper(tag_create_or_update_model, tag_update_model)


def _create_temp_alarm_tag(
        sensor: SensorModel,
        base_properties: Dict[str, Any],
        path_prefix: str,
        workspace: str,
        timestamp: datetime) -> Optional[TagWrapper]:
    enum_value = sensor.alarm
    value = SENSOR_HEALTH_STATE_VALUES.get(enum_value, None)
    if enum_value == SensorHealthState.UNKNOWN or value is None:
        return None
    optional_sensor_props_keys = [
        SensorProperties.UpperRisk,
        SensorProperties.LowerRisk,
        SensorProperties.UpperCritical,
        SensorProperties.LowerCritical,
        SensorProperties.UpperUser,
        SensorProperties.LowerUser,
    ]
    path_prefix = "{0}.{1}".format(path_prefix, ALARM)
    units = None
    tag_type = TYPE_INT
    display_name = _get_sensor_alarm_display_name(sensor, base_properties)
    props = {**base_properties,
             PROPS_ALARM_LEVEL: enum_value}
    return _create_sensor_tag(
        sensor=sensor,
        value=value,
        tag_type=tag_type,
        display_name=display_name,
        base_properties=props,
        path_prefix=path_prefix,
        optional_props_keys=optional_sensor_props_keys,
        units=units,
        workspace=workspace,
        timestamp=timestamp)


def _get_sensor_alarm_display_name(sensor: SensorModel, base_properties: Dict[str, str]) -> str:
    display_name = _get_sensor_reading_display_name(sensor.name, base_properties)
    return "{0} alarm".format(display_name)


def _create_temp_state_tag(
        sensor: SensorModel,
        base_properties: Dict[str, any],
        path_prefix: str,
        workspace: str,
        timestamp: datetime) -> Optional[TagWrapper]:
    enum_value = sensor.state
    value = SENSOR_HEALTH_STATE_VALUES.get(enum_value, None)
    if enum_value == SensorHealthState.UNKNOWN or value is None:
        return None
    optional_sensor_props_keys = [
        SensorProperties.UpperRisk,
        SensorProperties.LowerRisk,
        SensorProperties.UpperCritical,
        SensorProperties.LowerCritical,
        SensorProperties.UpperUser,
        SensorProperties.LowerUser,
        SensorProperties.Interval
    ]
    path_prefix = "{0}.{1}".format(path_prefix, STATE)
    units = None
    tag_type = TYPE_INT
    display_name = _get_sensor_state_display_name(sensor.name, base_properties)
    props = {**base_properties,
             PROPS_STATE_NAME: enum_value}
    return _create_sensor_tag(
        sensor=sensor,
        value=value,
        tag_type=tag_type,
        display_name=display_name,
        base_properties=props,
        path_prefix=path_prefix,
        optional_props_keys=optional_sensor_props_keys,
        units=units,
        workspace=workspace,
        timestamp=timestamp)


def _get_sensor_state_display_name(sensor_name: str, base_properties: Dict[str, str]) -> str:
    display_name = _get_sensor_reading_display_name(sensor_name, base_properties)
    return "{0} state".format(display_name)


def _create_volt_tags(
        volt_sensors_collection: SensorCollection,
        base_properties: Dict[str, str],
        path_prefix: str,
        workspace: str) -> List[TagWrapper]:
    path_prefix = "{0}.{1}".format(path_prefix, VOLTAGE)
    if not SensorCollection.has_any_readings(volt_sensors_collection):
        return []
    tags = []
    for sensor in volt_sensors_collection.sensors:
        tags.extend(_create_volt_sensor_tags(
            sensor=sensor,
            base_properties=base_properties,
            path_prefix=path_prefix,
            workspace=workspace,
            timestamp=volt_sensors_collection.timestamp))
    return tags


def _create_volt_sensor_tags(
        sensor: SensorModel,
        base_properties: Dict[str, str],
        path_prefix: str,
        workspace: str,
        timestamp: datetime) -> List[TagWrapper]:
    reading_tag = _create_volt_reading_tag(
        sensor=sensor,
        base_properties=base_properties,
        path_prefix=path_prefix,
        workspace=workspace,
        timestamp=timestamp)
    alarm_tag = _create_volt_alarm_tag(
        sensor=sensor,
        base_properties=base_properties,
        path_prefix=path_prefix,
        workspace=workspace,
        timestamp=timestamp)
    state_tag = _create_volt_state_tag(
        sensor=sensor,
        base_properties=base_properties,
        path_prefix=path_prefix,
        workspace=workspace,
        timestamp=timestamp)
    sensor_tags = [reading_tag, alarm_tag, state_tag]
    return [t for t in sensor_tags if t is not None]


def _create_volt_reading_tag(
        sensor: SensorModel,
        base_properties: Dict[str, str],
        path_prefix: str,
        workspace: str,
        timestamp: datetime) -> Optional[TagWrapper]:
    value = sensor.reading
    if value == -1:
        return None
    optional_sensor_props_keys = [
        SensorProperties.Nominal,
        SensorProperties.UpperRisk,
        SensorProperties.LowerRisk,
        SensorProperties.UpperCritical,
        SensorProperties.LowerCritical,
        SensorProperties.UpperUser,
        SensorProperties.LowerUser,
        SensorProperties.Accuracy
    ]
    path_prefix = "{0}.{1}".format(path_prefix, READING)
    units = UNITS_VOLTS
    tag_type = TYPE_DOUBLE
    display_name = _get_sensor_reading_display_name(sensor.name, base_properties)
    return _create_sensor_tag(
        sensor=sensor,
        value=value,
        tag_type=tag_type,
        display_name=display_name,
        base_properties=base_properties,
        path_prefix=path_prefix,
        optional_props_keys=optional_sensor_props_keys,
        units=units,
        workspace=workspace,
        timestamp=timestamp)


def _create_volt_alarm_tag(
        sensor: SensorModel,
        base_properties: Dict[str, str],
        path_prefix: str,
        workspace: str,
        timestamp: datetime) -> Optional[TagWrapper]:
    enum_value = sensor.alarm
    value = SENSOR_HEALTH_STATE_VALUES.get(enum_value, None)
    if enum_value == SensorHealthState.UNKNOWN or value is None:
        return None
    optional_sensor_props_keys = [
        SensorProperties.Nominal,
        SensorProperties.UpperRisk,
        SensorProperties.LowerRisk,
        SensorProperties.UpperCritical,
        SensorProperties.LowerCritical,
        SensorProperties.UpperUser,
        SensorProperties.LowerUser
    ]
    value = SENSOR_HEALTH_STATE_VALUES.get(enum_value, None)
    path_prefix = "{0}.{1}".format(path_prefix, ALARM)
    units = None
    tag_type = TYPE_INT
    display_name = _get_sensor_alarm_display_name(sensor, base_properties)
    props = {**base_properties,
             PROPS_ALARM_LEVEL: enum_value}
    return _create_sensor_tag(
        sensor=sensor,
        value=value,
        tag_type=tag_type,
        display_name=display_name,
        base_properties=props,
        path_prefix=path_prefix,
        optional_props_keys=optional_sensor_props_keys,
        units=units,
        workspace=workspace,
        timestamp=timestamp)


def _create_volt_state_tag(
        sensor: SensorModel,
        base_properties: Dict[str, str],
        path_prefix: str,
        workspace: str,
        timestamp: datetime) -> Optional[TagWrapper]:
    enum_value = sensor.state
    value = SENSOR_HEALTH_STATE_VALUES.get(enum_value, None)
    if enum_value == SensorHealthState.UNKNOWN or value is None:
        return None
    optional_sensor_props_keys = [
        SensorProperties.Nominal,
        SensorProperties.UpperRisk,
        SensorProperties.LowerRisk,
        SensorProperties.UpperCritical,
        SensorProperties.LowerCritical,
        SensorProperties.UpperUser,
        SensorProperties.LowerUser,
        SensorProperties.Interval
    ]
    value = SENSOR_HEALTH_STATE_VALUES.get(enum_value, None)
    path_prefix = "{0}.{1}".format(path_prefix, STATE)
    units = None
    tag_type = TYPE_INT
    display_name = _get_sensor_state_display_name(sensor.name, base_properties)
    props = {**base_properties,
             PROPS_STATE_NAME: enum_value}
    return _create_sensor_tag(
        sensor=sensor,
        value=value,
        tag_type=tag_type,
        display_name=display_name,
        base_properties=props,
        path_prefix=path_prefix,
        optional_props_keys=optional_sensor_props_keys,
        units=units,
        workspace=workspace,
        timestamp=timestamp)


def _create_power_tags(
        power_sensors_collection: SensorCollection,
        base_properties: Dict[str, str],
        path_prefix: str,
        workspace: str) -> List[TagWrapper]:
    path_prefix = "{0}.{1}".format(path_prefix, POWER)
    if not SensorCollection.has_any_readings(power_sensors_collection):
        return []
    tags = []
    for sensor in power_sensors_collection.sensors:
        tags.extend(_create_power_sensor_tags(
            sensor=sensor,
            base_properties=base_properties,
            path_prefix=path_prefix,
            workspace=workspace,
            timestamp=power_sensors_collection.timestamp))
    return tags


def _create_power_sensor_tags(
        sensor: SensorModel,
        base_properties: Dict[str, str],
        path_prefix: str,
        workspace: str,
        timestamp: datetime) -> List[TagWrapper]:
    reading_tag = _create_power_reading_tag(
        sensor=sensor,
        base_properties=base_properties,
        path_prefix=path_prefix,
        workspace=workspace,
        timestamp=timestamp)
    alarm_tag = _create_power_alarm_tag(
        sensor=sensor,
        base_properties=base_properties,
        path_prefix=path_prefix,
        workspace=workspace,
        timestamp=timestamp)
    state_tag = _create_power_state_tag(
        sensor=sensor,
        base_properties=base_properties,
        path_prefix=path_prefix,
        workspace=workspace,
        timestamp=timestamp)
    sensor_tags = [reading_tag, alarm_tag, state_tag]
    return [t for t in sensor_tags if t is not None]


def _create_power_reading_tag(
        sensor: SensorModel,
        base_properties: Dict[str, str],
        path_prefix: str,
        workspace: str,
        timestamp: datetime) -> Optional[TagWrapper]:
    value = sensor.reading
    if value == -1:
        return None
    optional_sensor_props_keys = [
        SensorProperties.UpperCritical,
        SensorProperties.UpperRisk,
        SensorProperties.UpperUser,
        SensorProperties.Accuracy
    ]
    path_prefix = "{0}.{1}".format(path_prefix, READING)
    units = UNITS_WATTS
    tag_type = TYPE_DOUBLE
    display_name = _get_sensor_reading_display_name(sensor.name, base_properties)
    return _create_sensor_tag(
        sensor=sensor,
        value=value,
        tag_type=tag_type,
        display_name=display_name,
        base_properties=base_properties,
        path_prefix=path_prefix,
        optional_props_keys=optional_sensor_props_keys,
        units=units,
        workspace=workspace,
        timestamp=timestamp)


def _create_power_alarm_tag(
        sensor: SensorModel,
        base_properties: Dict[str, str],
        path_prefix: str,
        workspace: str,
        timestamp: datetime) -> Optional[TagWrapper]:
    enum_value = sensor.alarm
    value = SENSOR_HEALTH_STATE_VALUES.get(enum_value, None)
    if enum_value == SensorHealthState.UNKNOWN or value is None:
        return None
    optional_sensor_props_keys = [
        SensorProperties.UpperCritical,
        SensorProperties.UpperRisk,
        SensorProperties.UpperUser,
    ]
    value = SENSOR_HEALTH_STATE_VALUES.get(enum_value, None)
    path_prefix = "{0}.{1}".format(path_prefix, ALARM)
    units = None
    tag_type = TYPE_INT
    display_name = _get_sensor_alarm_display_name(sensor, base_properties)
    props = {**base_properties,
             PROPS_ALARM_LEVEL: enum_value}
    return _create_sensor_tag(
        sensor=sensor,
        value=value,
        tag_type=tag_type,
        display_name=display_name,
        base_properties=props,
        path_prefix=path_prefix,
        optional_props_keys=optional_sensor_props_keys,
        units=units,
        workspace=workspace,
        timestamp=timestamp)


def _create_power_state_tag(
        sensor: SensorModel,
        base_properties: Dict[str, str],
        path_prefix: str,
        workspace: str,
        timestamp: datetime) -> Optional[TagWrapper]:
    enum_value = sensor.state
    value = SENSOR_HEALTH_STATE_VALUES.get(enum_value, None)
    if enum_value == SensorHealthState.UNKNOWN or value is None:
        return None
    optional_sensor_props_keys = [
        SensorProperties.UpperCritical,
        SensorProperties.UpperRisk,
        SensorProperties.UpperUser,
        SensorProperties.Interval
    ]
    value = SENSOR_HEALTH_STATE_VALUES.get(enum_value, None)
    path_prefix = "{0}.{1}".format(path_prefix, STATE)
    units = None
    tag_type = TYPE_INT
    display_name = _get_sensor_state_display_name(sensor.name, base_properties)
    props = {**base_properties,
             PROPS_STATE_NAME: enum_value}
    return _create_sensor_tag(
        sensor=sensor,
        value=value,
        tag_type=tag_type,
        display_name=display_name,
        base_properties=props,
        path_prefix=path_prefix,
        optional_props_keys=optional_sensor_props_keys,
        units=units,
        workspace=workspace,
        timestamp=timestamp)


def _create_lifetime_power_on_tags(
        lifetime_power_on: LifetimePowerOn,
        base_properties: Dict[str, str],
        path_prefix: str,
        workspace: str) -> List[TagWrapper]:
    if lifetime_power_on is None:
        return []
    tag = _create_lifetime_power_on_tag(
        hours=lifetime_power_on.power_on_hours,
        base_properties=base_properties,
        path_prefix=path_prefix,
        workspace=workspace,
        timestamp=lifetime_power_on.timestamp)
    if tag is not None:
        return [tag]
    return []


def _create_lifetime_power_on_tag(
        hours: int,
        base_properties: Dict[str, str],
        path_prefix: str,
        workspace: str,
        timestamp: datetime) -> TagWrapper:
    value = hours
    units = UNITS_HOURS
    tag_type = TYPE_INT
    display_name = _get_sensor_reading_display_name(LIFETIME_POWER_ON, base_properties)

    return _create_tag(
        name=LIFETIME_POWER_ON,
        value=value,
        tag_type=tag_type,
        display_name=display_name,
        base_properties=base_properties,
        path_prefix=path_prefix,
        units=units,
        workspace=workspace,
        timestamp=timestamp,
        collect_aggregates=False)


def _create_tag(
        name: str,
        value: Optional[Any],
        tag_type: str,
        display_name: str,
        base_properties: Dict[str, Any],
        path_prefix: str,
        units: Optional[str],
        workspace: str,
        timestamp: datetime,
        collect_aggregates: bool) -> Optional[TagWrapper]:
    if value is None or timestamp is None:
        return None
    value = str(value)
    sanitized_name = _sanitize_path_component(name)
    path = "{0}.{1}".format(path_prefix, sanitized_name)
    hyperlink = _get_hyperlink(path=path, workspace=workspace)
    props = {**base_properties,
             RESERVED_DISPLAY_NAME_KEY: display_name, RESERVED_HYPERLINK_KEY: hyperlink}
    if units is not None:
        props[RESERVED_UNITS_KEY] = units
    return _create_tag_wrapper(path, value, props, tag_type, timestamp, collect_aggregates)


def _create_operational_state_tags(
        operational_state: OperationalState,
        base_properties: Dict[str, str],
        path_prefix: str,
        workspace: str) -> List[TagWrapper]:
    if operational_state is None:
        return []
    if not operational_state.state or operational_state.state == OperationalStateValue.UNKNOWN:
        return []
    props = {**base_properties,
             PROPS_OPERATIONAL_STATE_STATUS: operational_state.status}
    tag = _create_operational_state_tag(
        state=operational_state.state,
        base_properties=props,
        path_prefix=path_prefix,
        workspace=workspace,
        timestamp=operational_state.timestamp)
    if tag is not None:
        return [tag]
    return []


def _create_operational_state_tag(
        state: str,
        base_properties: Dict[str, str],
        path_prefix: str,
        workspace: str,
        timestamp: datetime) -> TagWrapper:
    value = state
    tag_type = TYPE_STRING
    display_name = _get_sensor_reading_display_name(OPERATIONAL_STATE, base_properties)

    return _create_tag(
        name=OPERATIONAL_STATE,
        value=value,
        tag_type=tag_type,
        display_name=display_name,
        base_properties=base_properties,
        path_prefix=path_prefix,
        units=None,
        workspace=workspace,
        timestamp=timestamp,
        collect_aggregates=False)


def _sanitize_path_component(component: str) -> str:
    return component.replace('.', '\u2024')


def _create_cooling_tags(
        cooling_health: CoolingHealth,
        base_properties: Dict[str, str],
        path_prefix: str,
        workspace: str) -> List[TagWrapper]:
    path_prefix = "{0}.{1}".format(path_prefix, COOLING)

    tags = []
    if cooling_health.fan_mode is not None and cooling_health.supported_fan_modes is not None:
        tags.append(_create_fan_mode_tag(
            fan_mode=cooling_health.fan_mode,
            supported_fan_modes=cooling_health.supported_fan_modes,
            base_properties=base_properties,
            path_prefix=path_prefix,
            workspace=workspace,
            timestamp=cooling_health.timestamp))

        if cooling_health.fan_mode == FanMode.SAFE_MANUAL and cooling_health.user_desired_fan_speed is not None:
            tags.append(_create_user_defined_fan_speed_tag(
                user_desired_fan_speed=cooling_health.user_desired_fan_speed,
                manual_rpm_upper_bound=cooling_health.manual_rpm_upper_bound,
                manual_rpm_lower_bound=cooling_health.manual_rpm_lower_bound,
                base_properties=base_properties,
                path_prefix=path_prefix,
                workspace=workspace,
                timestamp=cooling_health.timestamp))

    if cooling_health.cooling_profile is not None and cooling_health.supported_cooling_profiles is not None:
        tags.append(_create_cooling_profile_tag(
            cooling_profile=cooling_health.cooling_profile,
            supported_cooling_profiles=cooling_health.supported_cooling_profiles,
            base_properties=base_properties,
            path_prefix=path_prefix,
            workspace=workspace,
            timestamp=cooling_health.timestamp))

    if cooling_health.cooling_operational_state is not None \
            and cooling_health.cooling_operational_state != OperationalStateValue.UNKNOWN:
        tags.append(_create_cooling_operational_state(
            cooling_operational_state=cooling_health.cooling_operational_state,
            base_properties=base_properties,
            path_prefix=path_prefix,
            workspace=workspace,
            timestamp=cooling_health.timestamp))

    if cooling_health.chassis_fans is not None:
        for sensor in cooling_health.chassis_fans:
            tags.extend(_create_fan_sensor_tags(
                sensor=sensor,
                base_properties=base_properties,
                path_prefix=path_prefix,
                workspace=workspace,
                timestamp=cooling_health.timestamp))

    if cooling_health.power_supply_fans is not None:
        for sensor in cooling_health.power_supply_fans:
            tags.extend(_create_power_supply_fan_sensor_tags(
                sensor=sensor,
                base_properties=base_properties,
                path_prefix=path_prefix,
                workspace=workspace,
                timestamp=cooling_health.timestamp))

    return tags


def _create_fan_mode_tag(
        fan_mode: str,
        supported_fan_modes: List[str],
        base_properties: Dict[str, str],
        path_prefix: str,
        workspace: str,
        timestamp: datetime) -> Optional[TagWrapper]:
    value = fan_mode
    tag_type = TYPE_STRING
    path_prefix = "{0}.{1}".format(path_prefix, FAN)
    display_name = _get_sensor_reading_display_name(FAN_MODE, base_properties)

    supported_fan_modes_property_value = ', '.join(supported_fan_modes)
    props = {**base_properties,
             PROPS_SUPPORTED_MODES: supported_fan_modes_property_value}
    return _create_tag(
        name=FAN_MODE,
        value=value,
        tag_type=tag_type,
        display_name=display_name,
        base_properties=props,
        path_prefix=path_prefix,
        units=None,
        workspace=workspace,
        timestamp=timestamp,
        collect_aggregates=False)


def _create_user_defined_fan_speed_tag(
        user_desired_fan_speed: int,
        manual_rpm_upper_bound: int,
        manual_rpm_lower_bound: int,
        base_properties: Dict[str, str],
        path_prefix: str,
        workspace: str,
        timestamp: datetime) -> Optional[TagWrapper]:
    value = user_desired_fan_speed
    units = UNITS_RPM
    tag_type = TYPE_INT
    display_name = _get_sensor_reading_display_name(FAN_USER_RPM, base_properties)
    path_prefix = "{0}.{1}".format(path_prefix, FAN)

    props = {**base_properties}
    if manual_rpm_upper_bound is not None:
        props = {**props,
                 PROPS_MANUAL_RPM_UPPER_BOUND: str(manual_rpm_upper_bound)}
    if manual_rpm_lower_bound is not None:
        props = {**props,
                 PROPS_MANUAL_RPM_LOWER_BOUND: str(manual_rpm_lower_bound)}
    return _create_tag(
        name=FAN_USER_RPM,
        value=value,
        tag_type=tag_type,
        display_name=display_name,
        base_properties=props,
        path_prefix=path_prefix,
        units=units,
        workspace=workspace,
        timestamp=timestamp,
        collect_aggregates=False)


def _create_cooling_profile_tag(
        cooling_profile: str,
        supported_cooling_profiles: List[str],
        base_properties: Dict[str, str],
        path_prefix: str,
        workspace: str,
        timestamp: datetime) -> Optional[TagWrapper]:

    # -1 is the numeric value considered to be "Custom"
    value = COOLING_HEALTH_COOLING_PROFILE_VALUES[cooling_profile]
    units = UNITS_WATTS if cooling_profile != CoolingProfile.PROFILE_CUSTOM else UNITS_CUSTOM
    tag_type = TYPE_INT
    display_name = _get_sensor_reading_display_name(COOLING_PROFILE, base_properties)

    converted_supported_cooling_profiles = []
    for supported_cooling_profile in supported_cooling_profiles:
        converted_cooling_profile = "{0} {1}".format(COOLING_HEALTH_COOLING_PROFILE_VALUES[supported_cooling_profile], UNITS_WATTS) \
            if supported_cooling_profile != CoolingProfile.PROFILE_CUSTOM \
            else "Custom"
        converted_supported_cooling_profiles.append(converted_cooling_profile)

    supported_cooling_profiles_property_value = ', '.join(converted_supported_cooling_profiles)
    props = {**base_properties,
             PROPS_SUPPORTED_MODES: supported_cooling_profiles_property_value}
    return _create_tag(
        name=COOLING_PROFILE,
        value=value,
        tag_type=tag_type,
        display_name=display_name,
        base_properties=props,
        path_prefix=path_prefix,
        units=units,
        workspace=workspace,
        timestamp=timestamp,
        collect_aggregates=False)


def _create_cooling_operational_state(
        cooling_operational_state: str,
        base_properties: Dict[str, str],
        path_prefix: str,
        workspace: str,
        timestamp: datetime) -> Optional[TagWrapper]:
    value = cooling_operational_state
    tag_type = TYPE_STRING
    display_name = _get_sensor_reading_display_name(COOLING_OPERATIONAL_STATE, base_properties)

    return _create_tag(
        name=COOLING_OPERATIONAL_STATE,
        value=value,
        tag_type=tag_type,
        display_name=display_name,
        base_properties=base_properties,
        path_prefix=path_prefix,
        units=None,
        workspace=workspace,
        timestamp=timestamp,
        collect_aggregates=False)


def _create_fan_sensor_tags(
        sensor: FanSensorModel,
        base_properties: Dict[str, str],
        path_prefix: str,
        workspace: str,
        timestamp: datetime):
    reading_tag = _create_fan_reading_tag(
        sensor=sensor,
        base_properties=base_properties,
        path_prefix=path_prefix,
        workspace=workspace,
        timestamp=timestamp
    )
    state_tag = _create_fan_state_tag(
        sensor=sensor,
        base_properties=base_properties,
        path_prefix=path_prefix,
        workspace=workspace,
        timestamp=timestamp
    )
    fan_tags = [reading_tag, state_tag]
    return [t for t in fan_tags if t is not None]


def _create_fan_reading_tag(
        sensor: FanSensorModel,
        base_properties: Dict[str, str],
        path_prefix: str,
        workspace: str,
        timestamp: datetime) -> Optional[TagWrapper]:
    value = sensor.reading
    if value == -1:
        return None
    path_prefix = "{0}.{1}.{2}".format(path_prefix, FAN, READING)
    units = UNITS_RPM
    tag_type = TYPE_DOUBLE
    display_name = _get_sensor_reading_display_name(sensor.name, base_properties)
    return _create_sensor_tag(
        sensor=sensor,
        value=value,
        tag_type=tag_type,
        display_name=display_name,
        base_properties=base_properties,
        path_prefix=path_prefix,
        optional_props_keys=[],
        units=units,
        workspace=workspace,
        timestamp=timestamp)


def _create_fan_state_tag(
        sensor: FanSensorModel,
        base_properties: Dict[str, str],
        path_prefix: str,
        workspace: str,
        timestamp: datetime) -> Optional[TagWrapper]:
    enum_value = sensor.state
    value = SENSOR_HEALTH_STATE_VALUES.get(enum_value, None)
    if enum_value == SensorHealthState.UNKNOWN or value is None:
        return None
    path_prefix = "{0}.{1}.{2}".format(path_prefix, FAN, STATE)
    units = None
    tag_type = TYPE_INT
    display_name = _get_sensor_state_display_name(sensor.name, base_properties)
    props = {**base_properties,
             PROPS_STATE_NAME: enum_value}
    return _create_sensor_tag(
        sensor=sensor,
        value=value,
        tag_type=tag_type,
        display_name=display_name,
        base_properties=props,
        path_prefix=path_prefix,
        optional_props_keys=[],
        units=units,
        workspace=workspace,
        timestamp=timestamp)


def _create_power_supply_fan_sensor_tags(
        sensor: FanSensorModel,
        base_properties: Dict[str, str],
        path_prefix: str,
        workspace: str,
        timestamp: datetime):
    path_prefix = "{0}.{1}".format(path_prefix, POWER)
    reading_tag = _create_fan_reading_tag(
        sensor=sensor,
        base_properties=base_properties,
        path_prefix=path_prefix,
        workspace=workspace,
        timestamp=timestamp
    )
    state_tag = _create_fan_state_tag(
        sensor=sensor,
        base_properties=base_properties,
        path_prefix=path_prefix,
        workspace=workspace,
        timestamp=timestamp
    )
    fan_tags = [reading_tag, state_tag]
    return [t for t in fan_tags if t is not None]


def _read_tag_last_updated_timestamp_file() -> Dict[str, str]:
    # Do not cast the timestamp strings to actual datetime objects. It is costly.
    filename = _get_tag_last_updated_timestamp_file_path()
    if not Path(filename).is_file():
        return {}
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    try:
        with open(filename, 'r') as f:
            return json.loads(f.read())
    except IOError as exc:
        log.error('Could not open tag timestamp file: ' + filename, exc, exc_info=True)
    except Exception as exc:
        log.error(
            'Unexpected exception in "_ni_asset_health_util._read_tag_last_updated_timestamp_file": %s',
            exc, exc_info=True
        )
    return {}


def _get_tag_last_updated_timestamp_file_path() -> str:
    return PathResolver.get_joined_skyline_appdata_dir(
        DATA_DIRECTORY,
        TAGS_DIRECTORY,
        LAST_UPDATED_TAGS_TIMESTAMPS_FILE)


def _update_tag_last_updated_timestamp(tag_path_to_timestamp: Dict[str, str]) -> None:
    stored_dict = _read_tag_last_updated_timestamp_file()
    stored_dict.update(tag_path_to_timestamp)
    _write_tag_last_updated_timestamp_file(stored_dict)


def _write_tag_last_updated_timestamp_file(timestamp_dict: Dict[str, str]) -> None:
    filename = _get_tag_last_updated_timestamp_file_path()
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    try:
        with open(filename, 'w') as f:
            return f.write(json.dumps(timestamp_dict, indent=2))
    except IOError as exc:
        log.error('Could not open tag timestamp file: ' + filename, exc, exc_info=True)
    except Exception as exc:
        log.error(
            'Unexpected exception in "_ni_asset_health_util._write_tag_last_updated_timestamp_file": %s',
            exc, exc_info=True
        )


def _get_tag_timestamp_from_tag_update(tag_wrapper: TagUpdate) -> str:
    return tag_wrapper.updates[0].timestamp.isoformat()


class TagHttpWriter:  # pylint: disable=too-few-public-methods
    """
    Abstraction over HTTP connection which facilitates communication
    between the minion and the Tag service.
    """

    def __init__(self) -> None:
        """
        Initialize the message service if not already initialized
        """
        self._configuration = None
        self._setup_communication()

    def get_configuration(self) -> Configuration:
        return self._configuration

    def _setup_communication(self) -> None:
        """
        Set up the communication.
        """
        file_path = get_http_master_file()
        if not os.path.isfile(file_path):
            # The Skyline Master file is not available.
            # Can't set up the Message Service without it.
            return

        self._setup_http_connection()

    def _setup_http_connection(self) -> None:
        """
        Set up the HTTP communication and initialize the Tag API object.
        """
        tag_route_name = '/nitag'
        self._configuration = get_configuration_by_id(
            HTTP_MASTER_CONFIGURATION_ID,
            tag_route_name,
            False)

    async def create_or_update_tags(
            self,
            tags_and_merge: TagListAndMergeFlag):
        """
        Publish an /update-tags request over HTTP.

        :param tags_and_merge: An instance of TagListAndMergeFlag representing the tags and the merge properties flag.
        """
        async with ApiClient(self._configuration) as api_client:
            tags_api = TagsApi(api_client=api_client)
            response = await tags_api.create_or_update_tags(tags_and_merge, _preload_content=False)
            if response.status not in (200, 201, 202):
                data = await response.text()
                rest_response = RESTResponse(response, data)
                raise ApiException(http_resp=rest_response)

    async def update_tag_values(self, updates: List[TagUpdate]) -> None:
        """
        Publish a /update-current-value request over HTTP.

        :param updates: A list of TagUpdate objects representing the new tag values.
        """
        async with ApiClient(self._configuration) as api_client:
            tags_api = TagsApi(api_client=api_client)
            response = await tags_api.update_tag_current_values(updates, _preload_content=False)
        if response.status not in (200, 202):
            data = await response.text()
            rest_response = RESTResponse(response, data)
            raise ApiException(http_resp=rest_response)
        else:
            tag_path_to_timestamp = {u.path: _get_tag_timestamp_from_tag_update(u)
                                     for u in updates if u.updates is not None}
            _update_tag_last_updated_timestamp(tag_path_to_timestamp)
