# coding: utf-8
'''
National Intruments SystemLink Assets Health Module
'''
import os
import sys
from typing import Any

# Import local libs
# This file may be loaded out of __pycache__, so the
# directory of its .py may not be in the search path.
IMPORT_PATH = os.path.dirname(__file__)
if IMPORT_PATH.endswith('__pycache__'):
    IMPORT_PATH = os.path.dirname(IMPORT_PATH)
sys.path.append(IMPORT_PATH)
try:
    import _nisysmgmt_grains
finally:
    # Remove the extra search path that we added to sys.path
    sys.path.remove(IMPORT_PATH)
# pylint: disable=import-error,3rd-party-module-not-gated

__virtualname__ = 'ni_asset_health'

DEFAULT_ASSET_HEALTH_MONITORING_ENABLED = True
DEFAULT_ASSET_HEALTH_MONITORING_INTERVAL = 300

MONITORING_ENABLED_KEY = 'asset_health_monitoring_enabled'
MONITORING_INTERVAL_KEY = 'asset_health_monitoring_interval'
HEALTH_BEACON_NAME = 'ni_asset_health_monitoring'

def __virtual__():
    return __virtualname__


def get_monitoring_enabled(**kwargs):
    '''
    Return whether or not asset health monitoring is enabled.

    :return: True if asset health monitoring is enabled, false otherwise.
    :rtype: bool
    '''
    return __grains__.get(MONITORING_ENABLED_KEY, DEFAULT_ASSET_HEALTH_MONITORING_ENABLED)


def get_monitoring_interval(**kwargs):
    '''
    Return the asset health monitoring interval.

    :return: The asset health monitoring interval in seconds
    :rtype: int
    '''
    return __grains__.get(MONITORING_INTERVAL_KEY, DEFAULT_ASSET_HEALTH_MONITORING_INTERVAL)


def _set_monitoring_option(key: str, value: Any) -> bool:
    if key in __grains__ and __grains__[key] == value:
        return False

    # This also writes the grain on disk
    __salt__['grains.setval'](key, value)
    # Avoid possible race condition with saltutil.refresh_grains
    _nisysmgmt_grains.refresh_grains_full(__opts__, __salt__)

    return True


def set_monitoring_enabled(enabled, **kwargs):
    '''
    Set whether or not asset health monitoring is enabled.

    :param enabled: Whether or not asset health monitoring is enabled.
    :type: bool
    '''
    enabled = True if enabled is True or enabled == 'True' else False
    changed = _set_monitoring_option(MONITORING_ENABLED_KEY, enabled)
    if changed:
        if enabled:
            func = 'enable_beacon'
        else:
            func = 'disable_beacon'
        __salt__['event.fire'](
            {'func': func, 'name': HEALTH_BEACON_NAME},
            'manage_beacons'
        )
    return True


def set_monitoring_interval(interval, **kwargs):
    '''
    Set the asset health monitoring interval.

    :param interval: The asset health monitoring interval in seconds.
    :type: int
    '''
    interval = int(interval)
    changed = _set_monitoring_option(MONITORING_INTERVAL_KEY, interval)
    if changed:
        __salt__['event.fire'](
            {
                'func': 'modify',
                'name': HEALTH_BEACON_NAME,
                'beacon_data': [{'interval': 1}]
            },
            'manage_beacons'
        )
    return True
